local _G = getfenv(0)
-- Local versions for performance
local string = _G.string
local math = _G.math
local pairs = _G.pairs
local ipairs = _G.ipairs
local table = _G.table
local floor, ceil, min, max, abs = math.floor, math.ceil, math.min, math.max, math.abs
local tconcat, tinsert = table.concat, table.insert
local wipe = _G.wipe
local tostring = _G.tostring
local tonumber = _G.tonumber

local CompactRunes = _G.LibStub("AceAddon-3.0"):NewAddon("Compact Runes", "AceConsole-3.0", "AceEvent-3.0","AceTimer-3.0")
local L = _G.LibStub("AceLocale-3.0"):GetLocale("CompactRunes", true)
local LDB = _G.LibStub("LibDataBroker-1.1")
local LibQTip = _G.LibStub("LibQTip-1.0")
local Icon = _G.LibStub("LibDBIcon-1.0")
local LSM = _G.LibStub:GetLibrary("LibSharedMedia-3.0")

-- Try to remove the Git hash at the end, otherwise return the passed in value.
local function cleanupVersion(version)
	local iter = string.gmatch(version, "(.*)-[a-z0-9]+$")
	if iter then
		local ver = iter()
		if ver and #ver >= 3 then
			return ver
		end
	end
	return version
end

local ADDON_NAME = ...
local ADDON_VERSION = cleanupVersion("0.8")
local options = nil
local MIN_UPDATE_TIME = 0.05

local function round(number)
    if not number then return 0 end
    return floor(number+0.5)
end

-- Define a simplistic class for combat statistics
local CombatStats = {}
CombatStats.__index = CombatStats

function CombatStats:new()
    local stats = {}
    _G.setmetatable(stats, CombatStats)
    stats:Reset()
    return stats
end

function CombatStats:Reset()
	self.rcProcs = 0
	self.cappedBC = 0
	self.reProcs = self.reProcs or {}
	wipe(self.reProcs)
	self.regenUses = self.regenUses or {}
	wipe(self.regenUses)
end

local RuneColors = {
	["BLOOD"] = { r = 0.77, g = 0.12, b = 0.23, a = 1.0 },
	["UNHOLY"] = { r = 0.3, g = 0.8, b = 0.1, a = 1.0 },
	["FROST"] = { r = 0.0, g = 0.4, b = 0.7, a = 1.0 },
	["DEATH"] = { r = 0.51, g = 0.23, b = 0.65, a = 1.0 },
}

local RuneCooldownColors = {
	["BLOOD"] = { r = 1.0, g = 0.0, b = 0.0, a = 0.1 },
	["UNHOLY"] = { r = 0.0, g = 1.0, b = 0.0, a = 0.1 },
	["FROST"] = { r = 0.0, g = 0.0, b = 1.0, a = 0.1 },
	["DEATH"] = { r = 1.0, g = 0.0, b = 1.0, a = 0.1 },
}

local RuneMapping = {
	[1] = "BLOOD",
	[2] = "UNHOLY",
	[3] = "FROST",
	[4] = "DEATH",
}

local HEXFMT = "|c%2x%2x%2x%2x"
local function chatColorCode(color)
	local a = 255
	local r = ceil(255 * color.r)
	local g = ceil(255 * color.g)
	local b = ceil(255 * color.b)
	return HEXFMT:format(a, r, g, b)
end

local RuneChatColors = {}
for k, v in pairs(RuneColors) do
	RuneChatColors[k] = chatColorCode(v)
end

-- Define Bar for now but the rest is at the bottom of the file.
local Bar = {}

CompactRunes.bars = {}
CompactRunes.playerName = UnitName("player")
CompactRunes.isDK = false
CompactRunes.TotalStats = CombatStats:new()
CompactRunes.LastFightStats = CombatStats:new()
CompactRunes.regenMethod = "None"
CompactRunes.currentSpec = ""
CompactRunes.currentProcs = {}
CompactRunes.soulReaperKnown = false
CompactRunes.currentTarget = nil
CompactRunes.targetSoulReaper = false
CompactRunes.targetHostile = false
CompactRunes.targetMaxHealth = 0
CompactRunes.srMethod = "cooldown"
CompactRunes.pillarReady = true
CompactRunes.baseFF = 0
CompactRunes.baseBP = 0
CompactRunes.effFF = 0
CompactRunes.effBP = 0
CompactRunes.FF = {}
CompactRunes.BP = {}
CompactRunes.cinderglacier = false
CompactRunes.effectiveAP = 0
CompactRunes.mastery = 0
CompactRunes.critChance = 0
CompactRunes.damageModifier = 1.0
CompactRunes.bloodCharges = 0
CompactRunes.bcExpires = 0
CompactRunes.MaxBloodCharges = 12
CompactRunes.BCPerUse = 2

local Broker = _G.CreateFrame("Frame")
Broker.obj = LDB:NewDataObject(_G.GetAddOnMetadata(ADDON_NAME, "Title"), {
    type = "data source",
    icon = "Interface\\Icons\\Spell_DeathKnight_RuneTap",
    label = _G.GetAddOnMetadata(ADDON_NAME, "Title"),
    text = _G.GetAddOnMetadata(ADDON_NAME, "Title"),
    barValue = 0,
    barR = 0,
    barG = 0,
    barB = 1,
	OnClick = function(clickedframe, button)
		if button == "RightButton" then
			local optionsFrame = _G.InterfaceOptionsFrame

			if optionsFrame:IsVisible() then
				optionsFrame:Hide()
			else
				CompactRunes:ShowOptions()
			end
		elseif button == "LeftButton" and _G.IsShiftKeyDown() then
		    CompactRunes.TotalStats:Reset()
		    CompactRunes.LastFightStats:Reset()
        end
	end
} )

local GREEN = "|cff00ff00"
local YELLOW = "|cffffff00"
local BLUE = "|cff0198e1"
local ORANGE = "|cffff9933"
local addonHdr = GREEN.."%s %s"
local totalDataHdr = ORANGE..L["Session"]
local lastFightValueHdr = ORANGE..L["Last Fight"]
local reProcFmt = "%s%d|r / %s%d|r / %s%d|r / %s%d"
local valuePlusPercFmt = "%d (%.1f%%)"
local reProcTypeHdr = "    %s%s"

local function valuePlusPercent(value, total)
	local val = value or 0
	local tot = total or 0
	local perc = (tot > 0) and (val / tot * 100) or 0
	return valuePlusPercFmt:format(val, perc)
end

local function AddStats(tooltip, stats)
	local totalProcs = 0
	local reProcs = stats.reProcs
	for k, v in pairs(reProcs) do
		totalProcs = totalProcs + (v or 0)
	end
	local reRegenUses = stats.regenUses["RE"] or 0
	local rcRegenUses = stats.regenUses["RC"] or 0
	local btRegenUses = stats.regenUses["BT"] or 0

    tooltip:AddSeparator(1)
	tooltip:AddLine(L["Blood Charges Capped"],
		valuePlusPercent(stats.cappedBC, btRegenUses))
	tooltip:AddLine(L["Runic Corruption Procs"],
		valuePlusPercent(stats.rcProcs, rcRegenUses))
	tooltip:AddLine(L["Runic Empowerment Procs"],
		valuePlusPercent(totalProcs, reRegenUses))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["BLOOD"],"Blood"),
			valuePlusPercent(reProcs[1], totalProcs))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["UNHOLY"],"Unholy"),
			valuePlusPercent(reProcs[2], totalProcs))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["FROST"],"Frost"),
			valuePlusPercent(reProcs[3], totalProcs))
	tooltip:AddLine(
		reProcTypeHdr:format(RuneChatColors["DEATH"],"Death"),
			valuePlusPercent(reProcs[4], totalProcs))


	--tooltip:AddLine("RE Procs By Type",
	--	reProcFmt:format(RuneChatColors["BLOOD"], (reProcs[1] or 0), 
	--		RuneChatColors["UNHOLY"], (reProcs[2] or 0), 
	--		RuneChatColors["FROST"], (reProcs[3] or 0), 
	--		RuneChatColors["DEATH"], (reProcs[4] or 0)))
end

function Broker.obj:OnEnter()
	local tooltip = LibQTip:Acquire("CompactRunesTooltip", 2, "LEFT", "RIGHT")
	self.tooltip = tooltip 

    tooltip:AddHeader(addonHdr:format(_G.GetAddOnMetadata(ADDON_NAME,"Title"), ADDON_VERSION))
    tooltip:AddLine()

    if CompactRunes.isDK then
        tooltip:AddLine(L["Shift + Left-Click to reset"], "", 1, 1, 1)
        tooltip:AddLine()

        tooltip:AddLine(totalDataHdr)
        AddStats(tooltip, CompactRunes.TotalStats)

        tooltip:AddLine()
	    tooltip:AddLine(lastFightValueHdr)
        AddStats(tooltip, CompactRunes.LastFightStats)	
    end

	tooltip:SmartAnchorTo(self)
	tooltip:Show()
end

function Broker.obj:OnLeave()
	LibQTip:Release(self.tooltip)
	self.tooltip = nil
end

local LookupOrKeyMT = {__index = function (t,k) return k end}
local SpellIds = {
	["Frost Fever"] = 59921,
	["Blood Plague"] = 59879,
	["Frost Fever Periodic"] = 55095,
	["Blood Plague Periodic"] = 55078,
	["Freezing Fog"] = 59052,
	["Killing Machine"] = 51128,
	["Sudden Doom"] = 49530,
	["Blood Charge"] = 114851,
	["Shadow Infusion"] = 91342,
	["Dark Transformation"] = 63560,
	["Death Coil"] = 47541,
	["Death Coil Damage"] = 47632,
	["Death Coil Heal"] = 47633,
	["Rune Strike"] = 56815,
	["Frost Strike"] = 49143,
	["Blood Tap"] = 45529,
	["Runic Empowerment"] = 81229,
	["Runic Corruption"] = 51460,
	["Dark Succor"] = 101568,
	["Crimson Scourge"] = 81141,
	["Plague Leech"] = 123693,
	["Plague Strike"] = 45462,
	["Outbreak"] = 77575,
	["Soul Reaper"] = 130735,
	["Soul Reaper (Blood)"] = 114866,
	["Soul Reaper (Frost)"] = 130735,
	["Soul Reaper (Unholy)"] = 130736,
	["Pillar of Frost"] = 51271,
	["Cinderglacier"] = 53386,
	["Auto Attack"] = 6603,
}
local SpellNames = {}
local SpellIcons = {}
_G.setmetatable(SpellNames, LookupOrKeyMT)
local function LoadSpellNames()
	for k, v in pairs(SpellIds) do
		local name, _, icon = _G.GetSpellInfo(v)
		SpellNames[k] = name
		SpellIcons[k] = icon
	end
end
LoadSpellNames()

local millFmt = "%.0fm"
local thousandFmt = "%.0fk"
local function FormatNumber(number)
    if tonumber(number) == nil then
        return number
    end

    if number > 1000000 then
        return millFmt:format(number / 1000000)
    elseif number > 1000 then
        return thousandFmt:format(number / 1000)
    end

    return number
end

local frames = {}

CompactRunes.PriorityListDefaults = {
	["Blood"] = {
		"Plague Leech",
		"Soul Reaper",
		"Crimson Scourge",					
		"Dark Succor",		
		"Blood Charge",
	},
	["Frost"] = {
		"Plague Leech",
		"Soul Reaper",
		"Killing Machine", 
		"Freezing Fog", 
		"Blood Charge",
		"Dark Succor",
		"Pillar of Frost",
	},
	["Unholy"] = {
		"Plague Leech",
		"Soul Reaper",
		"Sudden Doom", 
		"Blood Charge",
		"Dark Succor",
	},
}

-- These procs are handled differently.
CompactRunes.specialProcs = {
	["Soul Reaper"] = true,
	["Plague Leech"] = true,
	["Pillar of Frost"] = true,
}

CompactRunes.soulReaperIds = {
	["Blood"] = SpellIds["Soul Reaper (Blood)"],
	["Frost"] = SpellIds["Soul Reaper (Frost)"],
	["Unholy"] = SpellIds["Soul Reaper (Unholy)"]
}

CompactRunes.regenAbilities = {
	[SpellIds["Death Coil"]] = true,
	[SpellIds["Rune Strike"]] = true,
	[SpellIds["Frost Strike"]] = true,
	[SpellIds["Death Coil Damage"]] = true,
}

function CompactRunes:UpdatePriorityList(spec)
	local settings = self.db.profile.priorityList[spec]
	local list = {}
	local uniques = {}
	local added = {}
	for i, v in ipairs(self.PriorityListDefaults[spec]) do
		uniques[v] = true
	end
	if settings then
		for i, v in ipairs(settings) do
			if v and uniques[v] and not added[v] then
				tinsert(list, v)
				added[v] = true
			end
		end
	end
	for i, v in ipairs(self.PriorityListDefaults[spec]) do
		if v and not added[v] then
			tinsert(list, v)
			added[v] = true
		end
	end
	self.db.profile.priorityList[spec] = list
end

function CompactRunes:UpdatePriorityListSettings()
	for spec, list in pairs(self.PriorityListDefaults) do
		self:UpdatePriorityList(spec)
	end
end

function CompactRunes:MigrateSettings()
	local version = self.db.profile.profileVersion
	if version and version < 2 then
		for spec, list in pairs(self.db.profile.priorityList) do
			self.db.profile.priorityList[spec] = {}
		end
	end
	self.db.profile.profileVersion = 2
end

function CompactRunes:ResetPriorityList(spec)
	self.db.profile.priorityList[spec] = {}
	self:UpdatePriorityList(spec)
end

function CompactRunes:RemoveDefaults()
	for spec, list in pairs(self.PriorityListDefaults) do
		local matches = true
		local profile = self.db.profile.priorityList[spec]
		for i, v in ipairs(list) do
			if not (v and profile[i] and profile[i] == v) then
				matches = false
			end
		end
		if matches then
			self.db.profile.priorityList[spec] = {}
		end
	end
end

local defaults = {
	profile = {
		minimap = {
			hide = true,
		},
		debug = false,
		locked = false,
		ooc_alpha = 0.4,
		rechargeAlpha = 0.3,
		runeRechargeNums = false,
		almostReadyTime = 0,
		almostReadyAlpha = 0.6,
		rangeInd = true,
		plRefresh = 3,
		plOutbreak = true,
		trackDiseases = true,
		diseaseThreshold = 0.2,
		font_face = "Friz Quadrata TT",
		font_size = 12,
		fontFlags = {
			OUTLINE = true,
			THICKOUTLINE = false,
			MONOCHROME = false
		},
		texture = "Compact Runes",
		border = "Compact Runes",
		runeOrder = "BFU",
		diseaseOrder = "BF",
		width = 30,
		height = 10,
		spacing = 4,
		["frame"] = {
			x = 0,
			y = 0,
			color = { r = 0.1, g = 0.1, b = 0.1, a = 1.0 },
			bdcolor = { r = 0.5, g = 0.5, b = 0.5, a = 1.0 },
			rccolor = { r = 1.0, g = 1.0, b = 1.0, a = 0.7 },
		},
		["procIcon"] = {
			scale = 1.0,
			userPlaced = false,
			x = 0,
			y = 0,
		},
		["runes"] = {
			inset = 6,
		},
		procs = true,
		enabled = {
			["Blood"] = {
				["Blood Charge"] = false,
				["Soul Reaper"] = true,
				["Crimson Scourge"] = true,					
				["Dark Succor"] = true,
				["Plague Leech"] = false,
			},
			["Frost"] = {
				["Soul Reaper"] = true,
				["Freezing Fog"] = true, 
				["Killing Machine"] = true, 
				["Blood Charge"] = false,
				["Dark Succor"] = true,
				["Plague Leech"] = false,
			},
			["Unholy"] = {
				["Soul Reaper"] = true,
				["Sudden Doom"] = true, 
				["Blood Charge"] = false,
				["Dark Succor"] = true,
				["Plague Leech"] = false,
			},
		},
		priorityList = {
			["Blood"] = {},
			["Frost"] = {},
			["Unholy"] = {},
		},
		-- Bars
		bars = {
			-- Provide defaults for all bars.
			-- These are inherited if no bar or no value is set.
			['**'] = {
				textcolor = { r = 1.0, g = 1.0, b = 1.0, a = 1.0 },
			},
			["RunicPowerBar"] = {
				enabled = true,
				userPlaced = false,
				x = nil,
				y = nil,
				color = { r = 0.0, g = 0.82, b = 1.0, a = 1.0 },
				bgcolor = { r = 0.0, g = 0.205, b = 0.25, a = 1.0 },
				textcolor = { r = 0.66, g = 0.82, b = 1.0, a = 1.0 },
			},
			["BloodShieldBar"] = {
				enabled = false,
				userPlaced = false,
				x = nil,
				y = nil,
				color = { r = 1.0, g = 0.1, b = 0.1, a = 1.0 },
				bgcolor = { r = 0.5, g = 0.05, b = 0.05, a = 1.0 },
				textcolor = { r = 1.0, g = 0.85, b = 0.85, a = 1.0 },
			},
			["FrostFeverBar"] = {
				enabled = true,
				color = { r = 0.0, g = 0.45, b = 0.6, a = 1.0 },
				bgcolor = { r = 0.0, g = 0.1125, b = 0.15, a = 0.9 },
				textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
				reapplycolor = {
					[-1] = { r = 0.8, g = 0.1, b = 0.1, a = 1.0 },
					[1] = { r = 0.8, g = 0.8, b = 0.1, a = 1.0 },
					[2] = { r = 0.1, g = 0.8, b = 0.1, a = 1.0 },
				},
			},
			["BloodPlagueBar"] = {
				enabled = true,
				color = { r = 0.5, g = 0.0, b = 0.6, a = 1.0 },
				bgcolor = { r = 0.125, g = 0.0, b = 0.15, a = 0.9 },
				textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
				reapplycolor = {
					[-1] = { r = 0.8, g = 0.1, b = 0.1, a = 1.0 },
					[1] = { r = 0.8, g = 0.8, b = 0.1, a = 1.0 },
					[2] = { r = 0.1, g = 0.8, b = 0.1, a = 1.0 },
				},
			},
			["DarkTransformationBar"] = {
				enabled = true,
				color = { r = 0.0, g = 0.5, b = 0.0, a = 1.0 },
				bgcolor = { r = 0.125, g = 0.0, b = 0.15, a = 0.9 },
				textcolor = { r = 0.83, g = 0.66, b = 0.86, a = 1.0 },
				alt_color = { r = 0.5, g = 0.25, b = 0.0, a = 1.0 },
				alt_textcolor = { r = 0.8, g = 0.8, b = 0.8, a = 1.0 },
			},
			["BloodChargeBar"] = {
				enabled = true,
				color = { r = 1.0, g = 0.0, b = 0.0, a = 1 },
				bgcolor = { r = 0.65, g = 0.0, b = 0.0, a = 0.8 },
				textcolor = { r = 1.0, g = 0.9, b = 0.9, a = 1.0 },
			},
		},
	},
}

function CompactRunes:GetOptions()
    if not options then
        options = {
            type = "group",
            name = _G.GetAddOnMetadata(ADDON_NAME, "Title"),
            args = {
				core = self:GetGeneralOptions(),
				layout = self:GetLayoutOptions(),
				procs = self:GetProcsOptions()
            }
        }
		options.args.profile = _G.LibStub("AceDBOptions-3.0"):GetOptionsTable(self.db)
    end
    return options
end

function CompactRunes:ShowOptions()
	_G.InterfaceOptionsFrame_OpenToCategory(self.optionsFrame.Main)
end

function CompactRunes:GetGeneralOptions()
	local core = {
	    order = 1,
		name = L["General Options"],
		type = "group",
		args = {
		    description = {
		        order = 1,
		        type = "description",
		        name = L["CompactRunes_Desc"],
		    },
		    generalOptions = {
		        order = 100,
		        type = "header",
		        name = L["General Options"],
		    },
			locked = {
				name = L["Lock"],
				desc = L["LockDesc"],
				type = "toggle",
				order = 110,
				set = function(info, val)
				    self.db.profile.locked = val
					if frames.main then
						frames.main.lock = val
						frames.main:EnableMouse(not val)
					end
				end,
                get = function(info)
					return self.db.profile.locked
				end,
			},
    	    minimap = {
    			order = 120,
                name = L["Minimap Button"],
                desc = L["Toggle the minimap button"],
                type = "toggle",
                set = function(info,val)
                    	-- Reverse the value since the stored value is to hide it
                        self.db.profile.minimap.hide = not val
                    	if self.db.profile.minimap.hide then
                    		Icon:Hide("CompactRunesLDB")
                    	else
                    		Icon:Show("CompactRunesLDB")
                    	end
                      end,
                get = function(info)
            	        -- Reverse the value since the stored value is to hide it
                        return not self.db.profile.minimap.hide
                      end,
            },
			runeOrder = {
				order = 130,
				name = L["Rune Order"],
				desc = L["RuneOrder_Desc"],
				type = "select",
				values = {
				    ["BFU"] = L["Blood, Frost, Unholy"],
				    ["BUF"] = L["Blood, Unholy, Frost"],
				    ["FBU"] = L["Frost, Blood, Unholy"],
				    ["FUB"] = L["Frost, Unholy, Blood"],
				    ["UFB"] = L["Unholy, Frost, Blood"],
				    ["UBF"] = L["Unholy, Blood, Frost"],
				},
				set = function(info, val)
				    self.db.profile.runeOrder = val
					self:SetRuneOrder(val)
				end,
                get = function(info)
                    return self.db.profile.runeOrder
                end,
			},
			diseaseOrder = {
				order = 140,
				name = L["Disease Order"],
				desc = L["DiseaseOrder_Desc"],
				type = "select",
				values = {
				    ["FB"] = SpellNames["Frost Fever"]..", "..SpellNames["Blood Plague"],
				    ["BF"] = SpellNames["Blood Plague"]..", "..SpellNames["Frost Fever"],
				},
				set = function(info, val)
				    self.db.profile.diseaseOrder = val
					if frames.ffBar and frames.bpBar then
						frames.ffBar:SetPoint()
						frames.bpBar:SetPoint()
					end
				end,
                get = function(info)
                    return self.db.profile.diseaseOrder
                end,
			},
			ooc_alpha = {
				order = 150,
				name = L["OOC Alpha"],
				desc = L["OOCAlpha_Desc"],
				type = "range",
				min = 0,
				max = 1,
				step = 0.05,
				set = function(info, val)
				    self.db.profile.ooc_alpha = val
					if frames.main and not _G.InCombatLockdown() then
						frames.main:SetAlpha(self.db.profile.ooc_alpha)
					end
				end,
				get = function(info, val)
				    return self.db.profile.ooc_alpha
				end,					
			},
			rangeInd = {
				name = L["Range Indicator"],
				desc = L["RangeInd_Desc"],
				type = "toggle",
				order = 160,
				set = function(info, val)
				    self.db.profile.rangeInd = val
				end,
                get = function(info)
					return self.db.profile.rangeInd
				end,
			},
		    refreshingRunes = {
		        order = 200,
		        type = "header",
		        name = L["Recharging Runes"],
		    },
			rechargeAlpha = {
				order = 210,
				name = L["Recharge Alpha"],
				desc = L["RechargeAlpha_Desc"],
				type = "range",
				min = 0,
				max = 1,
				step = 0.05,
				set = function(info, val)
				    self.db.profile.rechargeAlpha = val
					if frames.main and not _G.InCombatLockdown() then
						frames.main:SetAlpha(self.db.profile.rechargeAlpha)
					end
				end,
				get = function(info, val)
				    return self.db.profile.rechargeAlpha
				end,					
			},
			showNums = {
				name = L["Show Timer"],
				desc = L["RechargeTimerDesc"],
				type = "toggle",
				order = 220,
				set = function(info, val)
				    self.db.profile.runeRechargeNums = val
					self:ShowRechargeNumbers(val)
				end,
                get = function(info)
					return self.db.profile.runeRechargeNums
				end,
			},
			almostReady = {
				order = 230,
				name = L["Almost Ready"],
				desc = L["AlmostReady_Desc"],
				type = "range",
				min = 0,
				max = 4,
				step = 0.5,
				set = function(info, val)
				    self.db.profile.almostReadyTime = val
				end,
				get = function(info, val)
				    return self.db.profile.almostReadyTime
				end,
			},
			almostReadyAlpha = {
				order = 240,
				name = L["Almost Ready Alpha"],
				desc = L["AlmostReadyAlpha_Desc"],
				type = "range",
				min = 0,
				max = 1,
				step = 0.05,
				set = function(info, val)
				    self.db.profile.almostReadyAlpha = val
				end,
				get = function(info, val)
				    return self.db.profile.almostReadyAlpha
				end,					
			},
		    features = {
		        order = 300,
		        type = "header",
		        name = L["Features"],
		    },
			runicpower = {
				name = L["Runic Power"],
				desc = L["RunicPowerDesc"],
				type = "toggle",
				order = 310,
				set = function(info, val)
				    self.db.profile.bars.RunicPowerBar.enabled = val
					self:ShowRunicPowerBar()
				end,
                get = function(info)
					return self.db.profile.bars.RunicPowerBar.enabled
				end,
			},
			frostFever = {
				name = SpellNames["Frost Fever"],
				desc = SpellNames["Frost Fever"],
				type = "toggle",
				order = 320,
				set = function(info, val)
				    self.db.profile.bars.FrostFeverBar.enabled = val
				end,
                get = function(info)
					return self.db.profile.bars.FrostFeverBar.enabled
				end,
			},
			bloodPlague = {
				name = SpellNames["Blood Plague"],
				desc = SpellNames["Blood Plague"],
				type = "toggle",
				order = 330,
				set = function(info, val)
				    self.db.profile.bars.BloodPlagueBar.enabled = val
				end,
                get = function(info)
					return self.db.profile.bars.BloodPlagueBar.enabled
				end,
			},
			darktrans = {
				name = SpellNames["Dark Transformation"],
				desc = SpellNames["Dark Transformation"],
				type = "toggle",
				order = 340,
				set = function(info, val)
				    self.db.profile.bars.DarkTransformationBar.enabled = val
					self:UpdateDTBCBars()
				end,
                get = function(info)
					return self.db.profile.bars.DarkTransformationBar.enabled
				end,
			},
			bloodcharge = {
				name = SpellNames["Blood Charge"],
				desc = SpellNames["Blood Charge"],
				type = "toggle",
				order = 350,
				set = function(info, val)
				    self.db.profile.bars.BloodChargeBar.enabled = val
					self:UpdateDTBCBars()
				end,
                get = function(info)
					return self.db.profile.bars.BloodChargeBar.enabled
				end,
			},
			bloodshield = {
				name = SpellNames["Blood Shield"],
				desc = SpellNames["Blood Shield"],
				type = "toggle",
				order = 360,
				set = function(info, val)
				    self.db.profile.bars.BloodShieldBar.enabled = val
				end,
                get = function(info)
					return self.db.profile.bars.BloodShieldBar.enabled
				end,
			},
			procs = {
				name = L["Procs"],
				desc = L["ProcsDesc"],
				type = "toggle",
				order = 370,
				set = function(info, val)
				    self.db.profile.procs = val
				end,
                get = function(info)
					return self.db.profile.procs
				end,
			},
			plagueLeech = {
		        order = 400,
		        type = "header",
		        name = SpellNames["Plague Leech"],
			},
			plRefreshTime = {
				order = 410,
				name = L["Refresh Time"],
				desc = L["PLRefresh_Desc"],
				type = "range",
				min = 1,
				max = 6,
				step = 0.5,
				set = function(info, val)
				    self.db.profile.plRefresh = val
				end,
				get = function(info, val)
				    return self.db.profile.plRefresh
				end,					
			},
			plOutbreak = {
				name = L["Require Outbreak"],
				desc = L["PLOutbreak_Desc"],
				type = "toggle",
				order = 420,
				set = function(info, val)
				    self.db.profile.plOutbreak = val
				end,
                get = function(info)
					return self.db.profile.plOutbreak
				end,
			},
			diseaseTrackingHdr = {
		        order = 450,
		        type = "header",
		        name = L["Disease Strength Tracking"],
			},
			trackDiseases = {
				name = L["Enabled"],
				desc = L["Enabled"],
				type = "toggle",
				order = 460,
				set = function(info, val)
				    self.db.profile.trackDiseases = val
				end,
                get = function(info)
					return self.db.profile.trackDiseases
				end,
			},
			diseaseThreshold = {
				order = 470,
				name = L["Threshold"],
				desc = L["Threshold"],
				type = "range",
				min = 0.01,
				max = 1,
				step = 0.01,
				bigStep = 0.05,
				isPercent = true,
				set = function(info, val)
				    self.db.profile.diseaseThreshold = val
				end,
				get = function(info, val)
				    return self.db.profile.diseaseThreshold
				end,
			},
		    dimensions = {
		        order = 500,
		        type = "header",
		        name = L["Dimensions"],
		    },
			width = {
				order = 510,
				name = L["Width"],
				desc = L["BarWidth_Desc"],	
				type = "range",
				min = 20,
				max = 100,
				step = 1,
				set = function(info, val)
				    self.db.profile.width = val
					self:Reset()
				end,
				get = function(info, val)
				    return self.db.profile.width
				end,
			},
			height = {
				order = 520,
				name = L["Height"],
				desc = L["BarHeight_Desc"],
				type = "range",
				min = 10,
				max = 50,
				step = 1,
				set = function(info, val)
				    self.db.profile.height = val
					self:Reset()
				end,
				get = function(info, val)
				    return self.db.profile.height
				end,					
			},
			procIconScale = {
				order = 530,
				name = L["Proc Indicator Scale"],
				desc = L["ProcIndScale_Desc"],
				type = "range",
				min = 1,
				max = 2,
				step = 0.01,
				bigStep = 0.05,
				isPercent = true,
				set = function(info, val)
				    self.db.profile.procIcon.scale = val
					frames.procInd:SetDimensions()
				end,
				get = function(info, val)
				    return self.db.profile.procIcon.scale
				end,					
			},
		    fonts = {
		        order = 600,
		        type = "header",
		        name = L["Font"],
		    },
			font_size = {
				order = 610,
				name = L["Font size"],
				desc = L["Font size for the bars."],
				type = "range",
				min = 8,
				max = 30,
				step = 1,
				set = function(info, val) 
					self.db.profile.font_size = val
					self:ResetFonts()
				end,
				get = function(info,val) return self.db.profile.font_size end,
			},
			font_face = {
				order = 620,
				type = "select",
				name = L["Font"],
				desc = L["Font to use."],
				values = LSM:HashTable("font"),
				dialogControl = 'LSM30_Font',
				get = function() return self.db.profile.font_face end,
				set = function(info, val) 
				    self.db.profile.font_face = val
					self:ResetFonts()
				end
			},
			font_outline = {
				name = L["Outline"],
				desc = L["FontOutlineDesc"],
				type = "toggle",
				order = 630,
				set = function(info, val)
				    self.db.profile.fontFlags.OUTLINE = val
				    self:ResetFonts()
				end,
                get = function(info)
                    return self.db.profile.fontFlags.OUTLINE
                end,
			},
			font_thickoutline = {
				name = L["Thick Outline"],
				desc = L["FontThickOutline_OptionDesc"],
				type = "toggle",
				order = 640,
				set = function(info, val)
				    self.db.profile.fontFlags.THICKOUTLINE = val
				    self:ResetFonts()
				end,
                get = function(info)
                    return self.db.profile.fontFlags.THICKOUTLINE
                end,
			},
			font_monochrome = {
				name = L["Monochrome"],
				desc = L["FontMonochromeDesc"],
				type = "toggle",
				order = 650,
				set = function(info, val)
				    self.db.profile.fontFlags.MONOCHROME = val
				    self:ResetFonts()
				end,
                get = function(info)
                    return self.db.profile.fontFlags.MONOCHROME
                end,
			},
            textures = {
                order = 700,
                type = "header",
                name = L["Textures"],
            },
			bartexture = {
				order = 710,
				name = L["Bar Texture"],
				desc = L["BarTexture_OptionDesc"],
				type = "select",
				values = LSM:HashTable("statusbar"),
				dialogControl = 'LSM30_Statusbar',
				get = function()
				    return self.db.profile.texture
				end,
				set = function(info, val)
				    self.db.profile.texture = val
					self:UpdateTextures()
				end,
			},
			bordertexture = {
				order = 720,
				name = L["Border"],
				desc = L["Border_OptionDesc"],
				type = "select",
				values = LSM:HashTable("border"),
				dialogControl = 'LSM30_Border',
				get = function()
				    return self.db.profile.border
				end,
				set = function(info, val)
				    self.db.profile.border = val
					self:SetMainFrameBorder()
				end,
			},
		},
	}
	return core
end

function CompactRunes:GetLayoutOptions()
	local layout = {
	    order = 1,
		name = L["Layout"],
		type = "group",
		inline = true,
		args = {
		    description = {
		        order = 1,
		        type = "description",
		        name = L["LayoutDesc"],
		    },
		    layoutHdr = {
		        order = 100,
		        type = "header",
		        name = L["Layout"],
		    },
		}
	}

	local bars = {
		["Proc Icon"] = {
			["frame"] = frames.procInd,
			["profile"] = self.db.profile.procIcon,
		},
		["Runic Power Bar"] = {
			["frame"] = frames.powerBar,
			["profile"] = self.db.profile.bars["RunicPowerBar"],
		},
		["Blood Shield Bar"] = {
			["frame"] = frames.bsBar,
			["profile"] = self.db.profile.bars["BloodShieldBar"],
		},
	}

	local i = 1
	for k, v in pairs(bars) do
		layout.args[k.."_Layout"] = {
			order = 101+i,
			type = "group",
			name = L[k],
			args = self:GetLayoutOptionsForBar(k, v),
		}
		i = i + 1
	end

	return layout
end

function CompactRunes:GetLayoutOptionsForBar(name, data)
	local frame = data.frame
	local profile = data.profile
	local layoutOpts = {
	    header = {
	        order = 100,
	        type = "header",
	        name = L[name],
	    },
		userPlaced = {
			name = L["Override"],
			desc = L["OverrideDesc"],
			type = "toggle",
			order = 110,
			set = function(info, val)
			    profile.userPlaced = val
				frame:SetDesiredPoint()
			end,
	        get = function(info)
				return profile.userPlaced
			end,
		},
		locked = {
			name = function()
				return frame:IsLocked() and L["Unlock"] or L["Lock"]
			end,
			desc = L["LockDesc"],
			type = "execute",
			order = 120,
			func = function(info, val)
				if frame then
					frame:Lock(not frame:IsLocked())
				end
			end,
			disabled = function()
				return not profile.userPlaced
			end,
		},
		procIconX = {
			order = 140,
			name = L["X Offset"],
			desc = L["XOffset_Desc"],	
			type = "range",
			softMin = -floor(_G.GetScreenWidth()),
			softMax = floor(_G.GetScreenWidth()),
			bigStep = 1,
			set = function(info, val)
			    profile.x = val
				frame:SetDesiredPoint()
			end,
			get = function(info, val)
			    return profile.x
			end,
			disabled = function()
				return profile.userPlaced == false
			end,
		},
		procIconY = {
			order = 150,
			name = L["Y Offset"],
			desc = L["YOffset_Desc"],	
			type = "range",
			softMin = -floor(_G.GetScreenHeight()),
			softMax = floor(_G.GetScreenHeight()),
			bigStep = 1,
			set = function(info, val)
			    profile.y = val
				frame:SetDesiredPoint()
			end,
			get = function(info, val)
			    return profile.y
			end,
			disabled = function()
				return profile.userPlaced == false
			end,
		},
	}
	return layoutOpts
end

function CompactRunes:GetProcsOptions()
	local procs = {
	    order = 2,
		name = L["Procs"],
		type = "group",
		inline = true,
		args = {
		    description = {
		        order = 1,
		        type = "description",
		        name = L["ProcsDesc"],
		    },
		    procsHdr = {
		        order = 100,
		        type = "header",
		        name = L["Procs"],
		    },
		},
	}

	local i = 1
	for k, v in pairs(self.db.profile.priorityList) do
		procs.args[k.."List"] = {
			order = 101+i,
			type = "group",
			name = k,
			args = self:GetProcsOptionsForSpec(k),
		}
		i = i + 1
	end

	return procs		
end

function CompactRunes:GetProcsOptionsForSpec(spec)
	local procList = {
        priorityList = {
	  	    order = 105,
	  	    type = "group",
	  	    inline = true,
	  	    name = L["Priority List"],
	  	    args = {
			},
		},
		reset = {
			order = 110,
			type = "execute",
			name = L["Reset Order"],
			confirmText = L["ResetOrderDesc"],
			confirm = true,
			func = function() self:ResetPriorityList(spec) end,
		},
	}
	
	if self.db.profile.priorityList[spec] then
		for k, v in pairs(self.db.profile.priorityList[spec]) do
			--local argName = v:gsub(" ", "_")
			procList.priorityList.args[v] = {
				name = SpellNames[v],
				desc = SpellNames[v],
				type = "toggle",
				order = function()
					local o = 0
					for pr, name in pairs(self.db.profile.priorityList[spec]) do
						if name == v then
							o = pr
						end
					end
					return 110 + (o * 5)
				end,
				set = function(info, val)
				    self.db.profile.enabled[spec][v] = val
				end,
		        get = function(info)
					return self.db.profile.enabled[spec][v]
				end,
			}
			procList.priorityList.args[v.."_Up"] = {
				name = L["Up"],
				desc = L["MoveUp_Desc"],
				type = "execute",
				width = "half",
				order = 111 + (k * 5),
				disabled = function()
					return k == 1
				end,
		        func = function(info)
					-- Swap with the one above it.
					local list = self.db.profile.priorityList[spec]
					local tmp = list[k - 1]

					if tmp then
						list[k - 1] = list[k]
						list[k] = tmp
					end
				end,
			}
			procList.priorityList.args[v.."_Down"] = {
				name = L["Down"],
				desc = L["MoveDown_Desc"],
				type = "execute",
				width = "half",
				order = 112 + (k * 5),
				disabled = function()
					return k >= #self.db.profile.priorityList[spec]
				end,
		        func = function(info)
					-- Swap with the one above it.
					local list = self.db.profile.priorityList[spec]
					local tmp = list[k + 1]
					if tmp then
						list[k + 1] = list[k]
						list[k] = tmp
					end
				end,
			}
		end
	end

	return procList
end

function CompactRunes:OnInitialize()
    -- Load the settings
    self.db = _G.LibStub("AceDB-3.0"):New("CompactRunesDB", defaults, "Default")

	self:MigrateSettings()
	self:UpdatePriorityListSettings()

	-- Register for profile callbacks
	self.db.RegisterCallback(self, "OnProfileChanged", "OnProfileChange")
	self.db.RegisterCallback(self, "OnProfileCopied", "OnProfileChange")
	self.db.RegisterCallback(self, "OnProfileReset", "OnProfileChange")
	self.db.RegisterCallback(self, "OnProfileShutdown", "RemoveDefaults")
	self.db.RegisterCallback(self, "OnDatabaseShutdown", "RemoveDefaults")
	
	LSM:Register("statusbar", "Compact Runes", [[Interface\Addons\CompactRunes\CompactRunes]])
	LSM:Register("border", "Compact Runes", [[Interface\Addons\CompactRunes\CompactRunesBorder]])
	LSM:Register("border", "Compact Runes Bright", [[Interface\Addons\CompactRunes\CompactRunesBrightBorder]])

	Icon:Register("CompactRunesLDB", Broker.obj, self.db.profile.minimap)
	LSM.RegisterCallback(CompactRunes, "LibSharedMedia_Registered")
end

function CompactRunes:OnProfileChange()
	self:MigrateSettings()
	self:UpdatePriorityListSettings()
	self:Reset()
end

function CompactRunes:Reset()
	--self:ResetFonts()
	--self:UpdateTextures()
	for k, v in pairs(self.bars) do
		if v then
			v.db = self.db.profile.bars[v.name]
			v:Reset()
		end
	end

	self:SetRuneOrder(self.db.profile.runeOrder)

	local procInd = frames.procInd
	if procInd then
		procInd.ResetFonts()
		procInd:SetWidth(procInd:GetDesiredWidth())
		procInd:SetHeight(procInd:GetDesiredHeight())
		procInd:SetDesiredPoint()
	end

	self:SetMainFrameBorder()
	local main = frames.main
	if main then
		main:SetWidth(main.GetDesiredWidth())
		main:SetHeight(main.GetDesiredHeight())
	end
end

function CompactRunes:LibSharedMedia_Registered(event, mediatype, key)
	if _G.strlen(self.db.profile.font_face) > 1 and mediatype == "font" then
		if self.db.profile.font_face == key then
			self:ResetFonts()
		end
	end
	if mediatype == "statusbar" then
	    self:UpdateTextures()
	end
end

function CompactRunes:ResetFonts()
	for k, v in pairs(self.bars) do
		if v then
			v:ResetFonts()
		end
	end

	if frames.procInd then
		frames.procInd.ResetFonts()
	end
end

function CompactRunes:UpdateTextures()
	for k, v in pairs(self.bars) do
		if v then
			v:UpdateTexture()
		end
	end
end

local function splitWords(str)
  local w = {}
  local function helper(word) table.insert(w, word) return nil end
  str:gsub("(%w+)", helper)
  return w
end

local function GetNumericValue(...)
    for i = 1, _G.select("#", ...) do
        local region = _G.select(i, ...)
        if region and region:GetObjectType() == "FontString" then
            local regionText = region:GetText()
            if regionText then
                local valueText = regionText:match("[0-9,]+")
                if valueText then
					valueText = valueText:gsub(",","")
                    local value = tonumber(valueText)
                    if value then
                        return value
                    end
                end
            end
        end
    end
end

local TipFrame = nil
function CompactRunes:VerifyDiseaseDamage()
	if not TipFrame then
		TipFrame = _G.CreateFrame("GameTooltip", "CompactRunes_TipFrame", nil, "GameTooltipTemplate")
		TipFrame:SetOwner(_G.WorldFrame, "ANCHOR_NONE")
	end

	TipFrame:ClearLines()
	TipFrame:SetSpellByID(SpellIds["Frost Fever"])
	local baseFF = GetNumericValue(TipFrame:GetRegions())
	TipFrame:ClearLines()
	TipFrame:SetSpellByID(SpellIds["Blood Plague"])
	local baseBP = GetNumericValue(TipFrame:GetRegions())

	self:CalculateDiseaseDamage()
	self:Print("Tooltip vs Calculated Values:")
	local fmt = "%s: Tooltip = %d,  Calculated = %d"
	self:Print(fmt:format("Frost Fever", baseFF, round(self.baseFF)))
	self:Print(fmt:format("Blood Plague", baseBP, round(self.baseBP)))
end

function CompactRunes:ChatCommand(input)
    if not input or input:trim() == "" then
        self:ShowOptions()
    else
		local cmds = splitWords(input)
        if cmds[1] and cmds[1] == "debug" then
			if cmds[2] and cmds[2] == "on" then
				self.db.profile.debug = true
	            self:Print("Debugging on.  Use '/crunes debug off' to disable.")
		    elseif cmds[2] and cmds[2] == "off" then
				self.db.profile.debug = false
	            self:Print("Debugging off.")
			else
				self:Print("Debugging is "..(self.db.profile.debug and "on." or "off."))
			end
		elseif cmds[1] and cmds[1] == "srMethod" then
			if cmds[2] and cmds[2] == "cooldown" then
				self.srMethod = "cooldown"
		    elseif cmds[2] and cmds[2] == "target" then
				self.srMethod = "target"
			end
			self:Print("Soul Reaper detected via " ..
				tostring(self.srMethod))
		elseif cmds[1] and cmds[1] == "verify" then
			self:VerifyDiseaseDamage()
		end
	end
end

function CompactRunes:OnEnable()
    self:CheckTalents()
	self:UNIT_ATTACK_POWER(nil, "player")
	self:COMBAT_RATING_UPDATE()
	self:UpdateDamageModifier()
	self:Load()
	LoadSpellNames()

	if not self.optionsFrame then
		-- Register Options
	    local displayName = _G.GetAddOnMetadata(ADDON_NAME, "Title")
		self:GetOptions()
	    _G.LibStub("AceConfigRegistry-3.0"):RegisterOptionsTable(displayName, options)

	    self.optionsFrame = {}
	    local ACD = _G.LibStub("AceConfigDialog-3.0")
		self.optionsFrame.Main = ACD:AddToBlizOptions(
		    displayName, displayName, nil, "core")
		self.optionsFrame.Layout = ACD:AddToBlizOptions(
		    displayName, L["Layout"], displayName, "layout")
		self.optionsFrame.Procs = ACD:AddToBlizOptions(
		    displayName, L["Procs"], displayName, "procs")
		ACD:AddToBlizOptions(
		    displayName, options.args.profile.name, displayName, "profile")
	    -- Register the chat command
	    self:RegisterChatCommand("crunes", "ChatCommand")
	end
end

local WatchedEvents = {
	"RUNE_POWER_UPDATE",
	"RUNE_TYPE_UPDATE",
	"UNIT_POWER_FREQUENT",
	"UNIT_MAXPOWER",
	"PLAYER_REGEN_DISABLED",
	"PLAYER_REGEN_ENABLED",
	"PLAYER_TARGET_CHANGED",
	"UNIT_AURA",
	"PLAYER_TALENT_UPDATE",
	"PLAYER_SPECIALIZATION_CHANGED",
	"COMBAT_LOG_EVENT_UNFILTERED",
	"PLAYER_ENTERING_WORLD",
	"UNIT_HEALTH",
	"UNIT_LEVEL",
	"SPELL_UPDATE_USABLE",
	"UNIT_ATTACK_POWER",
	"PLAYER_DAMAGE_DONE_MODS",
	--"UNIT_PET",
}
function CompactRunes:Load()
	if self.isDK then
		self:CreateDisplay()
		for i = 1, #WatchedEvents do
			self:RegisterEvent(WatchedEvents[i])
		end
	end
end

function CompactRunes:Unload()
end

function CompactRunes:OnDisable()
	for i = 1, #WatchedEvents do
		self:UnregisterEvent(WatchedEvents[i])
	end
end

function CompactRunes:CheckTalents()
    local class, className = _G.UnitClass("player")
    if className then
        if (className == 'DEATH KNIGHT' or className == 'DEATHKNIGHT') then
            self.isDK = true

			local activeSpecNum = _G.GetSpecialization()
			if activeSpecNum and activeSpecNum > 0 then
				local id, name, desc, texture = _G.GetSpecializationInfo(activeSpecNum)
		    	if texture == "Interface\\Icons\\Spell_Deathknight_BloodPresence" then
					self.currentSpec = "Blood"
				elseif texture == "Interface\\Icons\\Spell_Deathknight_FrostPresence" then
					self.currentSpec = "Frost"
				elseif texture == "Interface\\Icons\\Spell_Deathknight_UnholyPresence" then
					self.currentSpec = "Unholy"
				else
					self:Print("Error detecing player spec.")
					self.currentSpec = "Blood"
				end
			end

			local soulReaperId = self.soulReaperIds[self.currentSpec]
			self.soulReaperKnown = soulReaperId and 
				_G.IsSpellKnown(soulReaperId) or false

			local regenFound = false
			local plFound = false
			for i = 1, 18 do
				local name, iconTexture, tier, column, selected, available = 
					_G.GetTalentInfo(i, "player")
				if name and selected and available then
					if name == SpellNames["Blood Tap"] then
						self.regenMethod = "BT"
						regenFound = true
					elseif name == SpellNames["Runic Empowerment"] then
						self.regenMethod = "RE"
						regenFound = true
					elseif name == SpellNames["Runic Corruption"] then
						self.regenMethod = "RC"
						regenFound = true
					elseif name == SpellNames["Plague Leech"] then
						plFound = true
					end
				end
			end

			if not regenFound then
				self.regenMethod = "None"
			end
			if plFound then
				self.plagueLeech = true
			else
				self.plagueLeech = false
			end
        else
            self.isDK = false
			self.regenMethod = "None"
        end
    end

	self:UpdateDTBCBars()

	if self.db.profile.debug then
		local fmt = "CheckTalents [isDK=%s,spec=%s,regen=%s,SR=%s]"
		self:Print(
			fmt:format(tostring(self.isDK),
				self.currentSpec, self.regenMethod, 
				tostring(self.soulReaperKnown)))
	end
end

local function onUpdateRune(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.showing.timer = self.showing.timer + elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		self.lastUpdate = 0
		local profile = CompactRunes.db.profile
		self:SetValue(self.showing.timer)

		local remain = abs(self.showing.duration - self.showing.timer)
		local almostReadyTime = profile.almostReadyTime
		if almostReadyTime > 0 and almostReadyTime >= remain then
			self:SetAlpha(profile.almostReadyAlpha)
		end
		if profile.runeRechargeNums then
			if self.showing.start > 0 then
				self.value:SetText(tostring(round(remain)))
			else
				self.value:SetText("")
			end
		end
	end
end

local function onUpdateDisease(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.timer = self.timer - elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		self.lastUpdate = 0
		if self.active then
			if self.timer < 0 then
				self.timer = 0
				self.active = false
				if self.parent.name == "BloodPlagueBar" and
					CompactRunes.currentProcs["Plague Leech"] ~= nil then
					CompactRunes.currentProcs["Plague Leech"] = nil
					CompactRunes:UpdateProcIcon()
				end
			end
		
			self:Show()
			self:SetValue(self.timer)
			self.value:SetText(tostring(round(self.timer)))
			local db = CompactRunes.db.profile
			if self.parent.name == "BloodPlagueBar" then
				if self.timer <= db.plRefresh then
					if CompactRunes.currentProcs["Plague Leech"] == nil then
						local showPL = true
						if db.plOutbreak then
							local start, duration, enabled = 
								_G.GetSpellCooldown(SpellIds["Outbreak"])
							if duration > 0 then
								local timeLeft = duration - (_G.GetTime() - start)
								if timeLeft - self.timer >= 1 then
									showPL = false
								end
							end 
						end
						if showPL then
							CompactRunes.currentProcs["Plague Leech"] = 1
							CompactRunes:UpdateProcIcon()
						end
					end
				end
			end
		else
			self.value:SetText("0")
			self:Hide()
			self:SetScript("OnUpdate", nil)
			if self.parent.name == "BloodPlagueBar" and 
				CompactRunes.currentProcs["Plague Leech"] ~= nil then
				CompactRunes.currentProcs["Plague Leech"] = nil
				CompactRunes:UpdateProcIcon()
			end
		end
	end
end

local function onUpdateDarkTrans(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.timer = self.timer - elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		self.lastUpdate = 0
		if self.active then
			if self.timer < 0 then
				self.timer = 0
				self.active = false
				self:SetScript("OnUpdate", nil)
				self:Hide()
			else
				self:Show()
				self:SetValue(self.timer)
				if self.dt then
					self.value:SetText(tostring(round(self.timer)))
				end
			end
		else
			self:Hide()
		end
	end
end

local function onUpdateBloodCharge(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	self.timer = self.timer - elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		self.lastUpdate = 0
		if self.active then
			if self.timer < 0 then
				self.timer = 0
				self.active = false
				self:SetScript("OnUpdate", nil)
				self:Hide()
			else
				self:Show()
				self:SetValue(self.timer)
			end
		else
			self:Hide()
		end
	end
end

local function onUpdateInd(self, elapsed)
	self.lastUpdate = (self.lastUpdate or 0) + elapsed
	if self.lastUpdate >= MIN_UPDATE_TIME then
		local outOfRange = false
		if _G.UnitExists("target") and not _G.UnitIsDead("target") and
			_G.UnitCanAttack("player", "target") then
			if _G.IsSpellInRange(SpellNames["Plague Strike"], "target") == 0 then
				outOfRange = true
			end
		end

		if outOfRange ~= self.outOfRange then
			self.outOfRange = outOfRange
			self:UpdateInd()
		end

		self.lastUpdate = 0
	end
end

local fontFlags = {}
function CompactRunes:GetFontSettings()
	wipe(fontFlags)
	for k, v in pairs(self.db.profile.fontFlags) do
		if v then tinsert(fontFlags, k) end
	end
	return self.db.profile.font_face, self.db.profile.font_size, tconcat(fontFlags, ",")
end

-- Define a generic class for the bars
Bar.__index = Bar

function Bar:Create(name, fontAdj, width, height, setPoint, setValuePoint)
    local object = _G.setmetatable({}, Bar)
	object.name = name
	object.fontAdj = fontAdj
	
	local widthFunc = width
	if _G.type(width) ~= "function" then
		widthFunc = function() return width end
	end
	object.GetWidth = widthFunc
	local heightFunc = width
	if _G.type(height) ~= "function" then
		heightFunc = function() return height end
	end
	object.GetHeight = height

	object.SetPoint = setPoint
	object.SetDesiredPoint = setPoint
	object.SetValuePoint = setValuePoint
	object:Initialize()
	-- Add the bar to the addon's table of bars
	CompactRunes.bars[name] = object
	return object
end

function Bar:Initialize()
    local bar = _G.CreateFrame("StatusBar", "CompactRunes_"  .. self.name, frames.main)
	self.bar = bar
	bar.parent = self
	self.db = CompactRunes.db.profile.bars[self.name]
	self.altcolor = false
	self.isRuneBar = self.name:find("RuneBar%d")
	bar:SetScale(1)
    bar:SetOrientation("HORIZONTAL")
    bar:SetWidth(self.GetWidth())
    bar:SetHeight(self.GetHeight())
	local bt = LSM:Fetch("statusbar", CompactRunes.db.profile.texture)
    bar:SetStatusBarTexture(bt)
    bar:GetStatusBarTexture():SetHorizTile(false)
    bar:GetStatusBarTexture():SetVertTile(false)
	if not self.isRuneBar then
    	local bc = self.db.color
    	bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
	end
    bar.bg = bar:CreateTexture(nil, "BACKGROUND")
    bar.bg:SetTexture(bt)
    bar.bg:SetAllPoints(true)
	if not self.isRuneBar then
		local bgc = self.db.bgcolor
    	bar.bg:SetVertexColor(bgc.r, bgc.g, bgc.b, bgc.a)
	end

	local ff, fh, fflags = CompactRunes:GetFontSettings()
	local font = LSM:Fetch("font", ff)
    bar.value = bar:CreateFontString(nil, "OVERLAY")
    bar.value:SetPoint("CENTER")
    bar.value:SetFont(font, fh + self.fontAdj, fflags)
    bar.value:SetJustifyH("CENTER")
    --bar.value:SetShadowOffset(1, -1)
   	local tc = self.db.textcolor
   	bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	if self.isRuneBar and not CompactRunes.db.profile.runeRechargeNums then
		bar.value:Hide()
	end
	if self.name == "BloodShieldBar" then
		bar.value:SetText("0")
	end

	if self.SetPoint then self:SetPoint() end
	if self.SetValuePoint then
		self:SetValuePoint()
	else
	    bar.value:SetPoint("CENTER")
	end

	if self.name == "FrostFeverBar" or 
		self.name == "BloodPlagueBar" or 
		self.name == "DarkTransformationBar" or
		self.name == "BloodChargeBar" or
		self.name == "BloodShieldBar" then
		bar.active = false
		bar.timer = 0
		bar:Hide()
	end

	bar.locked = true
end

function Bar:IsLocked()
	return self.bar.locked
end

function Bar:Lock(locked)
	if locked ~= nil then
		self.bar.locked = locked
	end
	self.bar:EnableMouse(not self.bar.locked)
	if self.name == "BloodShieldBar" then
		if not self.bar.locked then
			self.bar:Show()
		elseif not self.bar.active then
			self.bar:Hide()
		end
	end
end

function Bar:SetMovable()
	local bar = self.bar
    bar:SetMovable()
    bar:RegisterForDrag("LeftButton")
    bar:SetScript("OnDragStart",
        function(self, button)
			if not self.locked then
            	self:StartMoving()
			end
        end)
    bar:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			self.parent.db.x, self.parent.db.y = x, y
			self:SetUserPlaced(false);
        end)
	self:Lock()
end

local numberFmt = "%.0f"
function Bar:SetValue(value)
	self.bar:SetValue(value)
	self.bar.value:SetText(numberFmt:format(value))
end

function Bar:SetValueTextColor(color)
   	local tc = color or self.db.textcolor
   	self.bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
end

function Bar:Reset()
	self:ResetFonts()
	self:UpdateVisibility()
	--self:UpdateTexture()
	self:UpdateGraphics()
end

function Bar:ResetFonts()
	local ff, fh, fontFlags = CompactRunes:GetFontSettings()
	local font = LSM:Fetch("font", ff)
	self.bar.value:SetFont(font, fh + self.fontAdj, fontFlags)						
	self.bar.value:SetText(self.bar.value:GetText())
end

function Bar:UpdateVisibility()
	if self.GetWidth then self.bar:SetWidth(self:GetWidth()) end
	if self.GetHeight then self.bar:SetHeight(self:GetHeight())	end
	if self.SetPoint and not self.isRuneBar then self:SetPoint() end
	if self.SetValuePoint then
		self:SetValuePoint()
	else
	    self.bar.value:SetPoint("CENTER")
	end

	self:UpdateTexture()
end

function Bar:UpdateTexture()
    local bt = LSM:Fetch("statusbar", CompactRunes.db.profile.texture)
	self.bar:SetStatusBarTexture(bt)
	self.bar.bg:SetTexture(bt)
    self.bar:GetStatusBarTexture():SetHorizTile(false)
    self.bar:GetStatusBarTexture():SetVertTile(false)
	--self:UpdateGraphics()
end

function Bar:UpdateGraphics()
	if not self.isRuneBar then
	    local bc, bgc, tc
		if self.altcolor then
		    bc = self.db.alt_color or self.db.color
		    bgc = self.db.alt_bgcolor or self.db.bgcolor
		    tc = self.db.alt_textcolor or self.db.textcolor
		else
		    bc = self.db.color
		    bgc = self.db.bgcolor
		    tc = self.db.textcolor
		end

	    self.bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
	    self.bar.bg:SetVertexColor(bgc.r, bgc.g, bgc.b, bgc.a)
	    self.bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	end
end

local runes = {
	[1] = {
		partner = 2,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[2] = {
		partner = 1,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[3] = {
		partner = 4,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[4] = {
		partner = 3,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[5] = {
		partner = 6,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
	[6] = {
		partner = 5,
		timer = 0,
		duration = 10,
		start = 0,
		type = 1,
		ready = true,
	},
}
local runeBars = {}
local orderings = {
	["BUF"] = { 1, 2, 3, 4, 5, 6 },
	["BFU"] = { 1, 2, 5, 6, 3, 4 },
	["UFB"] = { 3, 4, 5, 6, 1, 2 },
	["UBF"] = { 3, 4, 1, 2, 5, 6 },
	["FBU"] = { 5, 6, 1, 2, 3, 4 },
	["FUB"] = { 5, 6, 3, 4, 1, 2 },
}
function CompactRunes:SetRuneOrder(ordering)
	for i = 1, 6 do
		local barNum = orderings[ordering][i]
		local runeBar = runeBars[barNum]
		if runeBar then
			local bar = runeBar.bar
			bar:ClearAllPoints()
		    bar:SetPoint("TOPLEFT", frames.main, "TOPLEFT", 
				1 + self.db.profile.runes.inset + (self.db.profile.width + self.db.profile.spacing) * ((i - 1) % 2), 
				- 1 - self.db.profile.runes.inset - (self.db.profile.height + self.db.profile.spacing) * (ceil(i / 2) - 1))
		end
	end
end

function CompactRunes:ShowRechargeNumbers(enabled)
	for k, v in pairs(self.bars) do
		if v and v.isRuneBar and v.bar and v.bar.value then
			if enabled then
				v.bar.value:Show()
			else
				v.bar.value:Hide()
			end
		end
	end

	for i = 1, 6 do
		local runeBar = runeBars[i]
		if runeBar then
			local bar = runeBar.bar
			bar.value:Show()
		end
	end
end

function CompactRunes:CheckSoulReaper()
	if self.soulReaperKnown and 
		self.db.profile.enabled[self.currentSpec]["Soul Reaper"] then
		local health = _G.UnitHealth("target") or 0
		local percent = self.targetMaxHealth > 0 and 
			health / self.targetMaxHealth or 0

		local srActive = false
		if self.srMethod == "cooldown" then
			local start, duration, enabled = 
				_G.GetSpellCooldown(self.soulReaperIds[self.currentSpec])
			if duration > 0 then
				local timeLeft = duration - (_G.GetTime() - start)
				if timeLeft >= 1 then
					srActive = true
				end
			end 
		else
			srActive = self.targetSoulReaper
		end

		local old = self.currentProcs["Soul Reaper"]
		if self.targetHostile and percent < 0.35 and not srActive then
			self.currentProcs["Soul Reaper"] = 1
		else
			self.currentProcs["Soul Reaper"] = nil
		end
		if old ~= self.currentProcs["Soul Reaper"] then
			self:UpdateProcIcon()
		end
	end
end

function CompactRunes:UpdateRuneType(runeIndex)
	local partnerIndex = runes[runeIndex].partner
	local bar = runeBars[runeIndex]
	local type = _G.GetRuneType(runeIndex)
	runes[runeIndex].type = type

	local time = _G.GetTime()
	local firstIndex = min(runeIndex, partnerIndex)
	local secondIndex = max(runeIndex, partnerIndex)
	local first = runeBars[firstIndex]
	local second = runeBars[secondIndex]
	local sooner, later = runes[firstIndex], runes[secondIndex]
	if time - sooner.start < time - later.start then
		sooner, later = runes[secondIndex], runes[firstIndex]
	end

	local rc, rcc
	-- Update first
	rc = RuneColors[RuneMapping[sooner.type]]
	rcc = RuneCooldownColors[RuneMapping[sooner.type]]
    first.bar:SetStatusBarColor(rc.r, rc.g, rc.b, 1)
	first.bar.bg:SetVertexColor(rcc.r, rcc.g, rcc.b, rcc.a)
	first.bar.showing = sooner
	-- Update second
	rc = RuneColors[RuneMapping[later.type]]
	rcc = RuneCooldownColors[RuneMapping[later.type]]
    second.bar:SetStatusBarColor(rc.r, rc.g, rc.b, 1)
	second.bar.bg:SetVertexColor(rcc.r, rcc.g, rcc.b, rcc.a)
	second.bar.showing = later
end

function CompactRunes:UpdateRuneCooldown(runeIndex, isEnergized)
	local partnerIndex = runes[runeIndex].partner
	local start, duration, ready = _G.GetRuneCooldown(runeIndex)
	local type = _G.GetRuneType(runeIndex)
	local rune = runes[runeIndex]
	local bar = runeBars[runeIndex]
	rune.ready = ready

	if rune.ready and self.regenMethod == "RE" then
		local timeLeft = rune.timer - rune.duration
		if timeLeft > 0.5 then
			self.TotalStats.reProcs[type] = (self.TotalStats.reProcs[type] or 0) + 1
			self.LastFightStats.reProcs[type] = (self.LastFightStats.reProcs[type] or 0) + 1
			if self.db.profile.debug then
				self:Print("RE Proc - " .. (RuneMapping[type] or "Unknown"))
			end
		end
		--local refFmt = "Rune %d: %0.2f [%0.2f, %0.2f]"
		--self:Print(refFmt:format(runeIndex, timeLeft, 
		--	rune.duration, rune.timer))
	end

	rune.start = start
	if ready == false then
		rune.duration = duration
		rune.timer = _G.GetTime() - start
	else
		rune.timer = 10
		rune.duration = 10
	end

	local time = _G.GetTime()
	local firstIndex = min(runeIndex, partnerIndex)
	local secondIndex = max(runeIndex, partnerIndex)
	local first = runeBars[firstIndex]
	local second = runeBars[secondIndex]
	local sooner, later = runes[firstIndex], runes[secondIndex]
	if time - sooner.start < time - later.start then
		sooner, later = runes[secondIndex], runes[firstIndex]
	end

	-- Update first
	first.bar.showing = sooner
	first.bar:SetMinMaxValues(0, sooner.duration)
	first.bar:SetValue(sooner.timer)
	if sooner.ready == false then
		first.bar:SetScript("OnUpdate", onUpdateRune)
		first.bar:SetAlpha(self.db.profile.rechargeAlpha)
	else
		first.bar:SetScript("OnUpdate", nil)
		first.bar:SetAlpha(1.0)
		if CompactRunes.db.profile.runeRechargeNums then
			first.bar.value:SetText("")
		end
	end
	-- Update second
	second.bar.showing = later
	second.bar:SetMinMaxValues(0, later.duration)
	second.bar:SetValue(later.timer)
	if later.ready == false then
		second.bar:SetScript("OnUpdate", onUpdateRune)
		second.bar:SetAlpha(self.db.profile.rechargeAlpha)
	else
		second.bar:SetScript("OnUpdate", nil)
		second.bar:SetAlpha(1.0)
		if CompactRunes.db.profile.runeRechargeNums then
			second.bar.value:SetText("")
		end
	end

	--self:CheckSoulReaper()
end

function CompactRunes:CreateProcIndicator()
    local bar = _G.CreateFrame("Frame", "CompactRunes_Proc", frames.main)
	bar.db = self.db.profile
	frames.procInd = bar
	bar.currentProc = nil
	bar.procList = nil
	bar.SetDesiredPoint = function(bar)
		bar:ClearAllPoints()
		local profile = CompactRunes.db.profile
		local x, y = profile.procIcon.x, profile.procIcon.y
		if profile.procIcon.userPlaced and x ~= nil and y ~= nil then
	    	bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
		else
	    	bar:SetPoint("BOTTOM", frames.main, "TOP", 0, 
				profile.spacing * 7 + profile.height + 2)
		end
	end
	bar:SetDesiredPoint()
	bar:SetScale(1)
	bar.GetDesiredHeight = function()
		local scale = self.db.profile.procIcon.scale
		return ceil(scale*(frames.main:GetWidth()*frames.main:GetHeight()*0.233)^0.5)
	end
	bar.GetDesiredWidth = function()
		local scale = self.db.profile.procIcon.scale
		return ceil(scale*(frames.main:GetWidth()*frames.main:GetHeight()*0.233)^0.5)
	end
	bar.SetDimensions = function(bar)
	    bar:SetWidth(bar:GetDesiredHeight())
	    bar:SetHeight(bar:GetDesiredWidth())
	end
	bar:SetDimensions()
	bar.texture = bar:CreateTexture(nil, "ARTWORK")
    bar.texture:SetAllPoints(bar)
    bar.texture:SetVertexColor(1, 1, 1)
	local ff, fh, fflags = self:GetFontSettings()
	local font = LSM:Fetch("font", ff)
    bar.value = bar:CreateFontString(nil, "OVERLAY")
    bar.value:SetPoint("BOTTOMRIGHT", bar, "BOTTOMRIGHT", 0, -2)
    bar.value:SetFont(font, fh + 3, fflags)
    bar.value:SetJustifyH("CENTER")
	bar.value:SetJustifyV("CENTER")
    bar.value:SetShadowOffset(1, -1)
    local tc = bar.db.bars.ProcInd.textcolor
    bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
	bar.value:Hide()
	
	bar.ResetFonts = function()
		local ff, fh, fontFlags = self:GetFontSettings()
		local font = LSM:Fetch("font", ff)
		bar.value:SetFont(font, fh + 3, fontFlags)						
		bar.value:SetText(bar.value:GetText())
	end

	bar.locked = true
	bar.Lock = function(self, locked)
		if locked ~= nil then
			self.locked = locked
		end
		self:EnableMouse(not self.locked)
		self:Configure(not self.locked)
	end
	bar.IsLocked = function(self)
		return self.locked
	end

	bar.configMode = false
	bar.Configure = function(self, config)
		if config ~= nil then
			self.configMode = config
		else
			self.configMode = not self.configMode
		end
		if self.configMode then
			self.texture:SetTexture(SpellIcons["Auto Attack"])
			self:Show()
		else
			CompactRunes:UpdateProcIcon()
		end
	end

    bar:SetMovable()
    bar:RegisterForDrag("LeftButton")
    bar:SetScript("OnDragStart",
        function(self, button)
			if not self.locked then
            	self:StartMoving()
			end
        end)
    bar:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			local procIcon = CompactRunes.db.profile.procIcon
			procIcon.x, procIcon.y = x, y
			procIcon.userPlaced = true
			self:SetUserPlaced(false)
        end)
	bar:Lock()
end

function CompactRunes:UpdateProcIcon()
	local procInd = frames.procInd
	local procPriority = self.db.profile.priorityList[self.currentSpec]
	local enabled = self.db.profile.enabled[self.currentSpec]
	local found = false

	if procPriority and enabled then
		for i, spell in ipairs(procPriority) do
			local count = self.currentProcs[spell]
			local isPL = spell == "Plague Leech"
			if enabled[spell] and count and 
				(not isPL or (isPL and self.plagueLeech)) then
				found = true
				procInd.texture:SetTexture(SpellIcons[spell])
				procInd:Show()
				if count > 1 then
					procInd.value:SetText(tostring(count))
					procInd.value:Show()
				else
					procInd.value:Hide()
				end

				break
			end
		end
	end

	if not found then
		procInd:Hide()
	end
end

function CompactRunes:SetMainFrameBorder()
	local frame = frames.main
	if frame then
		local border = LSM:Fetch("border", self.db.profile.border)
		local backdrop = {
	        bgFile = "Interface\\Buttons\\WHITE8X8", 
	        edgeFile = border, tile = false, tileSize = 16, edgeSize = 16, 
	        insets = { left = 4, right = 4, top = 4, bottom = 4 }
	    }
		local bdbc = self.db.profile.frame.bdcolor
		local bdc = self.db.profile.frame.color
	    frame:SetBackdrop(backdrop)
	    frame:SetBackdropBorderColor(bdbc.r, bdbc.g, bdbc.b, bdbc.a)
	    frame:SetBackdropColor(bdc.r, bdc.g, bdc.b, bdc.a)

		border = LSM:Fetch("border", "Compact Runes Bright")
		backdrop = {
	        bgFile = "Interface\\Buttons\\WHITE8X8", 
	        edgeFile = border, tile = false, tileSize = 16, edgeSize = 16, 
	        insets = { left = 4, right = 4, top = 4, bottom = 4 }
	    }
		local rc = self.db.profile.frame.rccolor
	    frame.ind:SetBackdrop(backdrop)
	    frame.ind:SetBackdropBorderColor(rc.r, rc.g, rc.b, rc.a)
	    frame.ind:SetBackdropColor(0, 0, 0, 0)
	end
end

function CompactRunes:IsBothDTBCBars()
	return self.currentSpec == "Unholy" and 
			self.regenMethod == "BT" and
			self.db.profile.bars.BloodChargeBar.enabled and
			self.db.profile.bars.DarkTransformationBar.enabled
end

function CompactRunes:UpdateDTBCBars()
	if self.isDK then
		local bars = {
			"DarkTransformationBar",
			"BloodChargeBar"
		}
		for k, v in pairs(bars) do
			if v and self.bars[v] then
				local bar = self.bars[v]
				bar:SetPoint()
				if not bar.db.enabled then
					bar.bar.active = false 
					bar.bar.timer = 0
					bar.bar:SetScript("OnUpdate", nil)
					bar.bar:Hide()
				end
			end
		end
	end
end

function CompactRunes:CalculateSize()
	self.sizes = self.sizes or {}
	self.sizes.Display = self.sizes.Display or {}
	self.sizes.Display.width = (self.db.profile.width + self.db.profile.spacing) * 2 + 9
	self.sizes.Display.height = (self.db.profile.height + self.db.profile.spacing) * 3 + 9
	self.sizes.hinset = 0
	self.sizes.vinset = 0
end

function CompactRunes:CreateDisplay()
	local frame = _G.CreateFrame("Frame", "CompactRunes_Display", _G.UIParent)
	frames.main = frame
	frame.db = self.db.profile.frame
	frame:SetPoint("CENTER", _G.UIParent, "CENTER", frame.db.x, frame.db.y)
	frame.GetDesiredWidth = function()
		return (self.db.profile.width + self.db.profile.spacing) * 2 + 10
	end
	frame.GetDesiredHeight = function()
		return (self.db.profile.height + self.db.profile.spacing) * 3 + 10
	end
	frame:SetWidth(frame.GetDesiredWidth())
	frame:SetHeight(frame.GetDesiredHeight())

	frame.ind = _G.CreateFrame("Frame", nil,frame)
	frame.ind:SetAllPoints(frame)
	frame.ind:Hide()
	frame.rcActive = false
	frame.outOfRange = false
	frame.UpdateInd = function(self)
		if self.outOfRange then
			self.ind:SetBackdropBorderColor(1, 0, 0, 0.6)
			self.ind:Show()
		elseif self.rcActive then
			self.ind:SetBackdropBorderColor(1, 1, 1, 0.6)
			self.ind:Show()
		else
			self.ind:Hide()
		end
	end
	self:SetMainFrameBorder()

	frame.lock = self.db.profile.locked
    frame:SetMovable()
    frame:RegisterForDrag("LeftButton")
    frame:SetScript("OnDragStart",
        function(self, button)
			if not self.lock then
            	self:StartMoving()
			end
        end)
    frame:SetScript("OnDragStop",
        function(self)
            self:StopMovingOrSizing()
			local scale = self:GetEffectiveScale() / _G.UIParent:GetEffectiveScale()
			local x, y = self:GetCenter()
			x, y = x * scale, y * scale
			x = x - _G.GetScreenWidth()/2
			y = y - _G.GetScreenHeight()/2
			x = x / self:GetScale()
			y = y / self:GetScale()
			self.db.x, self.db.y = x, y
			self:SetUserPlaced(false);
        end)
    frame:EnableMouse(not frame.lock)

	frames.powerBar = Bar:Create("RunicPowerBar", 2, 
		function() return self.db.profile.width * 2 + self.db.profile.spacing end,
		function() return self.db.profile.height end,
		function(self)
			self.bar:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self.bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				self.bar:SetPoint("TOPLEFT", frames.main, "BOTTOMLEFT", 
					CompactRunes.db.profile.runes.inset + 1, -2)
			end
		end
	)
	frames.powerBar:SetMovable()
	self:UpdateRunicPowerMax()
	self:UpdateRunicPower()
	self:ShowRunicPowerBar()

	frames.bsBar = Bar:Create("BloodShieldBar", 2, 
		function() return self.db.profile.width * 2 + self.db.profile.spacing end,
		function() return self.db.profile.height end,
		function(self)
			self.bar:ClearAllPoints()
			local x, y = self.db.x, self.db.y
			if self.db.userPlaced and x ~= nil and y ~= nil then
				self.bar:SetPoint("CENTER", _G.UIParent, "CENTER", x, y)
			else
				self.bar:SetPoint("TOPLEFT", frames.main, "BOTTOMLEFT", 
					CompactRunes.db.profile.runes.inset + 1, 
					-2 - CompactRunes.db.profile.height - 6)
			end
		end
	)
	frames.bsBar:SetMovable()

	frames.ffBar = Bar:Create("FrostFeverBar", 2, 
		function() return self.db.profile.width end,
		function() return self.db.profile.height end,
		function(self)
			local db = CompactRunes.db.profile
			self.bar:ClearAllPoints()
			if db.diseaseOrder == "BF" then
				self.bar:SetPoint("BOTTOMRIGHT", frames.main, "TOPRIGHT", 
					-db.runes.inset, 0)
			else
				self.bar:SetPoint("BOTTOMLEFT", frames.main, "TOPLEFT", 
					db.runes.inset, 0)
			end
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)
	frames.ffBar.reapply = nil

	frames.bpBar = Bar:Create("BloodPlagueBar", 2, 
		function() return self.db.profile.width end,
		function() return self.db.profile.height end,
		function(self)
			local db = CompactRunes.db.profile
			self.bar:ClearAllPoints()
			if db.diseaseOrder == "FB" then
				self.bar:SetPoint("BOTTOMRIGHT", frames.main, "TOPRIGHT", 
					-db.runes.inset, 0)
			else
				self.bar:SetPoint("BOTTOMLEFT", frames.main, "TOPLEFT", 
					db.runes.inset, 0)
			end
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)
	frames.bpBar.reapply = nil

	frames.dtBar = Bar:Create("DarkTransformationBar", 2, 
		function() return self.db.profile.width end,
		function() return self.db.profile.height end,
		function(self)
			self.bar:ClearAllPoints()
			local profile = CompactRunes.db.profile
			local yOffset = profile.spacing * 2 + profile.height + 2
			if CompactRunes:IsBothDTBCBars() then
				self.bar:SetPoint("BOTTOMLEFT", frames.main, "TOPLEFT", 
					profile.runes.inset, yOffset)
			else
				self.bar:SetPoint("BOTTOM", frames.main, "TOP", 
					0, yOffset)
			end
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)

	frames.bcBar = Bar:Create("BloodChargeBar", 2, 
		function() return self.db.profile.width end,
		function() return self.db.profile.height end,
		function(self)
			self.bar:ClearAllPoints()
			local profile = CompactRunes.db.profile
			local yOffset = profile.spacing * 2 + profile.height + 2
			if CompactRunes:IsBothDTBCBars() then
				self.bar:SetPoint("BOTTOMRIGHT", frames.main, "TOPRIGHT", 
					-profile.runes.inset, yOffset)
			else
				self.bar:SetPoint("BOTTOM", frames.main, "TOP", 
					0, yOffset)
			end
		end,
		function(self) self.bar.value:SetPoint("CENTER", self.bar, "TOP", 0, 0) end
	)

	self:CreateProcIndicator()
	
	-- Create the runes and then update their status
	for i = 1, 6 do
		runeBars[i] = Bar:Create("RuneBar"..i, 0, 
			function() return self.db.profile.width end,
			function() return self.db.profile.height end,
			nil,
			nil
		)
		runeBars[i].bar.showing = runes[i]
	end
	for i = 1, 6 do
		self:UpdateRuneType(i)
		self:UpdateRuneCooldown(i)
	end	
	self:SetRuneOrder(self.db.profile.runeOrder)
	frames.main:SetAlpha(self.db.profile.ooc_alpha)
	frames.main:Show()
end

function CompactRunes:ShowRunicPowerBar()
	if self.db.profile.bars.RunicPowerBar.enabled then
		frames.powerBar.bar:Show()
		self:RegisterEvent("UNIT_POWER_FREQUENT")
	else
		frames.powerBar.bar:Hide()
		self:UnregisterEvent("UNIT_POWER_FREQUENT")
	end
end

function CompactRunes:RUNE_POWER_UPDATE(event, runeIndex, isEnergized)
	self:UpdateRuneCooldown(runeIndex, isEnergized)
end

function CompactRunes:RUNE_TYPE_UPDATE(event, runeIndex)
	self:UpdateRuneType(runeIndex)
end

function CompactRunes:UNIT_POWER_FREQUENT(event, unit, type)
	if unit == "player" then -- and type == "RUNIC_POWER" then
		self:UpdateRunicPower()
	end
end

function CompactRunes:UpdateRunicPower()
	frames.powerBar:SetValue(_G.UnitPower("player", 6))
end

function CompactRunes:UpdateRunicPowerMax()
	frames.powerBar.bar:SetMinMaxValues(0, _G.UnitPowerMax("player", 6))
end

-- Frost Vulnerability from Razorice is dynamic.  No need to calculate it.
-- 5.1 FF: 143 + 0.138 * AP
-- 5.1 BP: 172 + 0.138 * AP
local BASEFF = 166
local BASEBP = 197
local APMOD = 0.158
function CompactRunes:CalculateDiseaseDamage()
	local baseFF = (BASEFF + APMOD * self.effectiveAP) * self.damageModifier
	local baseBP = (BASEBP + APMOD * self.effectiveAP) * self.damageModifier

	if self.currentSpec == "Frost" then
		baseFF = baseFF * (1 + (2.0 * self.mastery / 100))
	elseif self.currentSpec == "Unholy" then
		baseFF = baseFF * 1.6
		baseBP = baseBP * (1 + (2.5 * self.mastery / 100)) * 1.6
	end

	local effFF = baseFF * (1 + (self.critChance / 100))
	local effBP = baseBP * (1 + (self.critChance / 100))

	if self.cinderglacier then
		effFF = effFF * 1.2
		effBP = effBP * 1.2
	end

	self.baseFF = baseFF
	self.baseBP = baseBP
	self.effFF = effFF
	self.effBP = effBP

	if self.db.profile.debug then
		local fmt = "FF: %d / %d, BP: %d/ %d"
		self:Print(fmt:format(
			round(baseFF), round(effFF), round(baseBP), round(effBP)))
	end

	self:CheckDiseases()
end

local reapplyFmt = "Reapply %s [%d %s %d : %.1f]"
function CompactRunes:CheckDiseases()
	if self.currentTarget and self.db.profile.trackDiseases then
		local snapshotFF = self.FF[self.currentTarget] or 0
		local threshold = self.db.profile.diseaseThreshold
		local reapplyFF = 0
		if snapshotFF > 0 then
			if self.effFF > snapshotFF then
				if self.effFF > snapshotFF * (1 + threshold) then
					reapplyFF = 2
				else
					reapplyFF = 1
				end
			elseif self.effFF < snapshotFF then
				reapplyFF = -1
			end
			if self.db.profile.debug and reapplyFF ~= 0 then
				self:Print(reapplyFmt:format(
					"FF", round(self.effFF), 
					reapplyFF == -1 and "<" or ">",
					round(snapshotFF),
					(self.effFF-snapshotFF)/snapshotFF*100))
			end
		end
		local ffBar = frames.ffBar
		if reapplyFF ~= ffBar.reapply then
			ffBar:SetValueTextColor(ffBar.db.reapplycolor[reapplyFF])
			ffBar.reapply = reapplyFF
		end

		local reapplyBP = 0
		local snapshotBP = self.BP[self.currentTarget] or 0
		if snapshotBP > 0 then
			if self.effBP > snapshotBP then
				if self.effBP > snapshotBP * (1 + threshold) then
					reapplyBP = 2
				else
					reapplyBP = 1
				end
			elseif self.effBP < snapshotBP then
				reapplyBP = -1
			end
			if self.db.profile.debug and reapplyBP ~= 0 then
				self:Print(reapplyFmt:format(
					"BP", round(self.effBP), 
					reapplyBP == -1 and "<" or ">",
					round(snapshotBP),
					(self.effBP-snapshotBP)/snapshotBP*100))
			end
		end
		local bpBar = frames.bpBar
		if reapplyBP ~= bpBar.reapply then
			bpBar:SetValueTextColor(bpBar.db.reapplycolor[reapplyBP])
			bpBar.reapply = reapplyBP
		end
	end
end

function CompactRunes:UNIT_MAXPOWER(event)
	self:UpdateRunicPowerMax()
end

function CompactRunes:UNIT_HEALTH(event, unit)
	if unit == "target" then
		self:CheckSoulReaper()
	end
end

function CompactRunes:UNIT_LEVEL(event, unit)
	if unit == "player" then
		self:CheckTalents()
	end
end

function CompactRunes:UNIT_ATTACK_POWER(event, unit)
	if unit == "player" then
		local base, posBuff, negBuff = _G.UnitAttackPower("player")
		self.effectiveAP = base + posBuff - negBuff
		self:CalculateDiseaseDamage()
	end
end

function CompactRunes:PLAYER_DAMAGE_DONE_MODS(event, unit)
	if unit == "player" then
		self:UpdateDamageModifier()
	end
end

function CompactRunes:UpdateDamageModifier()
	self.damageModifier = _G.select(7, _G.UnitDamage("player")) or 1.0
	self:CalculateDiseaseDamage()
end

function CompactRunes:COMBAT_RATING_UPDATE()
	self.mastery = _G.GetMastery()
	self.critChance = _G.GetCritChance()
	self:CalculateDiseaseDamage()
end

function CompactRunes:PLAYER_REGEN_DISABLED()
	-- Player has entered combat.
	frames.main:SetAlpha(1)
	self.LastFightStats:Reset()
	if self.db.profile.rangeInd then
		frames.main:SetScript("OnUpdate", onUpdateInd)
	end
end

function CompactRunes:PLAYER_REGEN_ENABLED()
	-- Player has left combat.
	frames.main:SetAlpha(self.db.profile.ooc_alpha)
	if self.db.profile.rangeInd and not self.targetHostile then
		frames.main:SetScript("OnUpdate", nil)
		frames.main.outOfRange = false
		frames.main:UpdateInd()
	end
end

function CompactRunes:PLAYER_TARGET_CHANGED(unit)
	self.targetHostile = _G.UnitExists("target") and 
		_G.UnitCanAttack("player", "target") and
		not _G.UnitIsDead("target")
	if self.targetHostile and self.db.profile.rangeInd then
		frames.main:SetScript("OnUpdate", onUpdateInd)
	else
		frames.main:SetScript("OnUpdate", nil)
		frames.main.outOfRange = false
		frames.main:UpdateInd()
	end

	self.targetMaxHealth = _G.UnitHealthMax("target") or 0
	self.currentTarget = _G.UnitGUID("target")

	if self.currentProcs["Plague Leech"] ~= nil then
		self.currentProcs["Plague Leech"] = nil
		self:UpdateProcIcon()
	end
	self:UNIT_AURA("UNIT_AURA", "target")
	self:CheckSoulReaper()
	frames.ffBar.reapply = nil
	frames.bpBar.reapply = nil
	self:CheckDiseases()
end

local petBuffs = {
	SpellNames["Dark Transformation"],
	SpellNames["Shadow Infusion"],
}
local diseases
function CompactRunes:UNIT_AURA(event, unit)
	local name, rank, icon, count, dispelType, duration, expires, 
		caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
		isBossDebuff, castByPlayer, value1, value2, value3 

	if unit == "target" then
		diseases = diseases or {
			[SpellNames["Frost Fever"]] = frames.ffBar,
			[SpellNames["Blood Plague"]] = frames.bpBar,
		}

		for spell, bar in pairs(diseases) do
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
			= _G.UnitDebuff("target", spell, nil, "PLAYER")

			if name and bar.db.enabled then
				bar.bar.active = true
				local timeLeft = expires - _G.GetTime()
				if timeLeft > bar.bar.timer + 1 then
					if self.currentProcs["Plague Leech"] ~= nil then
						self.currentProcs["Plague Leech"] = nil
						self:UpdateProcIcon()
					end
				end
				bar.bar.timer = timeLeft
				bar.bar:SetMinMaxValues(0, duration)
				bar.bar:Show()
				bar.bar:SetScript("OnUpdate", onUpdateDisease)
			else
				bar.bar.active = false 
				bar.bar.timer = 0
				bar.bar:SetScript("OnUpdate", nil)
				bar.bar:Hide()
				if spell == "Blood Plague" then
					if self.currentProcs["Plague Leech"] ~= nil then
						self.currentProcs["Plague Leech"] = nil
						self:UpdateProcIcon()
					end
				end
			end
		end

		name, rank, icon, count, dispelType, duration, expires, 
		caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
		isBossDebuff, castByPlayer, value1, value2, value3 
		= _G.UnitDebuff("target", SpellNames["Soul Reaper"], nil, "PLAYER")
		local old = self.targetSoulReaper
		if name then
			self.targetSoulReaper = true
		else
			self.targetSoulReaper = false
		end
		if old ~= self.targetSoulReaper then
			self:UpdateProcIcon()
		end
	elseif unit == "player" then
		if self.db.profile.bars.BloodShieldBar.enabled then
			local bsBar = frames.bsBar
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
				= _G.UnitBuff("player", SpellNames["Blood Shield"])
			if name then
				bsBar.bar.timer = expires - _G.GetTime()
				bsBar.bar:SetMinMaxValues(0, duration)
				bsBar.bar.value:SetText(FormatNumber(value1 or 0))
				bsBar.bar.active = true 
				bsBar.bar:Show()
				bsBar.bar:SetScript("OnUpdate", onUpdateBloodCharge)
			else
				bsBar.bar.active = false 
				bsBar.bar.timer = 0
				bsBar.bar:SetScript("OnUpdate", nil)
				bsBar.bar:Hide()
			end
		end

		if self.db.profile.bars.BloodChargeBar.enabled then
			local bcBar = frames.bcBar
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
				= _G.UnitBuff("player", SpellNames["Blood Charge"])
			if name then
				if expires > self.bcExpires then
					if self.bloodCharges >= self.MaxBloodCharges - 1 then
						self.TotalStats.cappedBC = self.TotalStats.cappedBC + 1
						self.LastFightStats.cappedBC = self.LastFightStats.cappedBC + 1
					end
				end
				self.bloodCharges = count
				self.bcExpires = expires
				bcBar.bar.timer = expires - _G.GetTime()
				bcBar.bar:SetMinMaxValues(0, duration)
				bcBar.bar.value:SetText(tostring(count))
				bcBar.bar.active = true 
				bcBar.bar:Show()
				bcBar.bar:SetScript("OnUpdate", onUpdateBloodCharge)
			else
				self.bloodCharges = 0
				self.bcExpires = 0
				bcBar.bar.active = false 
				bcBar.bar.timer = 0
				bcBar.bar:SetScript("OnUpdate", nil)
				bcBar.bar:Hide()
			end
		end

		if self.db.profile.procs then
			local procInd = frames.procInd
			local procPriority = self.db.profile.priorityList[self.currentSpec]
			local enabled = self.db.profile.enabled[self.currentSpec]
			if procPriority and enabled then
				for i, spell in ipairs(procPriority) do
					if not self.specialProcs[spell] then
						name, rank, icon, count, dispelType, duration, expires, 
						caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
						isBossDebuff, castByPlayer, value1, value2, value3 
						= _G.UnitBuff("player", SpellNames[spell])
						if name then
							self.currentProcs[spell] = (count or 1)
						else
							self.currentProcs[spell] = nil
						end
					end
				end
			end
			self:UpdateProcIcon()
		end

		if self.db.profile.trackDiseases then
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
				= _G.UnitBuff("player", SpellNames["Cinderglacier"])
			local cinder = (name ~= nil)
			if cinder ~= self.cinderglacier then
				self:CalculateDiseaseDamage()
				self.cinderglacier = cinder
			end
		end

	elseif unit == "pet" then
		for i, spell in pairs(petBuffs) do
			name, rank, icon, count, dispelType, duration, expires, 
			caster, isStealable, shouldConsolidate, spellId, canApplyAura, 
			isBossDebuff, castByPlayer, value1, value2, value3 
			= _G.UnitBuff("pet", spell, nil, "PLAYER")
			if name then break end
		end

		local dtBar = frames.dtBar
		if name then
			dtBar.bar.dt = (name == SpellNames["Dark Transformation"])
			dtBar.altcolor = dtBar.bar.dt
			dtBar.bar.timer = expires - _G.GetTime()
			local dtdb = self.db.profile.bars.DarkTransformationBar
			local bc, tc
			if dtBar.altcolor then
				bc = dtdb.alt_color
				tc = dtdb.alt_textcolor
			else
				bc = dtdb.color
				tc = dtdb.textcolor
				dtBar.bar.value:SetText(tostring(count))
			end
			dtBar.bar:SetMinMaxValues(0, duration)
			dtBar.bar:SetStatusBarColor(bc.r, bc.g, bc.b, bc.a)
			dtBar.bar.value:SetTextColor(tc.r, tc.g, tc.b, tc.a)
			if self.db.profile.bars.DarkTransformationBar.enabled then
				dtBar.bar.active = true 
				dtBar.bar:Show()
				dtBar.bar:SetScript("OnUpdate", onUpdateDarkTrans)
			end
		else
			dtBar.bar.active = false 
			dtBar.bar.timer = 0
			dtBar.bar:SetScript("OnUpdate", nil)
			dtBar.bar:Hide()
		end
	end
end

function CompactRunes:PLAYER_SPECIALIZATION_CHANGED(event)
	self:CheckTalents()
end

function CompactRunes:PLAYER_TALENT_UPDATE(event)
	self:CheckTalents()
end

function CompactRunes:COMBAT_LOG_EVENT_UNFILTERED(...)
    local event, timestamp, eventtype, hideCaster, 
        srcGUID, srcName, srcFlags, srcRaidFlags, 
        destGUID, destName, destFlags, destRaidFlags, 
        param9, param10, param11, param12, param13, param14, 
        param15, param16, param17, param18, param19, param20

    event, timestamp, eventtype, hideCaster, 
    srcGUID, srcName, srcFlags, srcRaidFlags,
    destGUID, destName, destFlags, destRaidFlags,
    param9, param10, param11, param12, param13, param14, 
    param15, param16, param17, param18, param19, param20 = ...

	if destGUID == self.playerGUID then
		if eventtype == "SPELL_AURA_APPLIED" then
			if param9 == SpellIds["Runic Corruption"] then
				self.LastFightStats.rcProcs = self.LastFightStats.rcProcs + 1
				self.TotalStats.rcProcs = self.TotalStats.rcProcs + 1
				if frames.main then
					frames.main.rcActive = true
					frames.main:UpdateInd()
				end
			end
		elseif eventtype == "SPELL_AURA_REMOVED" then
			if param9 == SpellIds["Runic Corruption"] then
				if frames.main then
					frames.main.rcActive = false
					frames.main:UpdateInd()
				end
			end
		end
	end
	if srcGUID == self.playerGUID then
		if eventtype == "SPELL_AURA_APPLIED" or eventtype == "SPELL_AURA_REFRESH" then
			if param9 == SpellIds["Frost Fever"] or param9 == SpellIds["Frost Fever Periodic"] then
				self.FF[destGUID] = self.effFF
				if destGUID == self.currentTarget then
					frames.ffBar:SetValueTextColor()
					frames.ffBar.reapply = nil
				end
			elseif param9 == SpellIds["Blood Plague"] or param9 == SpellIds["Blood Plague Periodic"] then
				self.BP[destGUID] = self.effBP
				if destGUID == self.currentTarget then
					frames.bpBar:SetValueTextColor()
					frames.bpBar.reapply = nil

					-- If Plague Leech proc is active, remove it.
					if self.currentProcs["Plague Leech"] ~= nil then
						self.currentProcs["Plague Leech"] = nil
						self:UpdateProcIcon()
					end
				end
			end
		elseif eventtype == "SPELL_AURA_REMOVED" then
			if param9 == SpellIds["Frost Fever"] or param9 == SpellIds["Frost Fever Periodic"] then
				self.FF[destGUID] = nil
				frames.ffBar.reapply = nil
			elseif param9 == SpellIds["Blood Plague"] or param9 == SpellIds["Blood Plague Periodic"] then
				self.BP[destGUID] = nil
				frames.bpBar.reapply = nil
				
				if destGUID == self.currentTarget then
					-- If Plague Leech proc is active, remove it.
					if self.currentProcs["Plague Leech"] ~= nil then
						self.currentProcs["Plague Leech"] = nil
						self:UpdateProcIcon()
					end
				end
			end
		elseif eventtype == "SPELL_DAMAGE" or 
			(eventtype == "SPELL_HEAL" and self.regenMethod == "BT") then
			local damage = param12 or 0
			if damage > 0 and self.regenAbilities[param9] ~= nil then
				local regenMethod = self.regenMethod
				self.TotalStats.regenUses[regenMethod] = (self.TotalStats.regenUses[regenMethod] or 0) + 1
				self.LastFightStats.regenUses[regenMethod] = (self.LastFightStats.regenUses[regenMethod] or 0) + 1
			end
		end
	end
end

function CompactRunes:SPELL_UPDATE_USABLE(event)
	local changed = false
	if self.currentSpec == "Frost" then
		local start, duration, enabled = 
			_G.GetSpellCooldown(SpellIds["Pillar of Frost"])
		local pillarReady = (duration == 0)
		self.currentProcs["Pillar of Frost"] = 
			(pillarReady and 1 or nil)
		if self.pillarReady ~= pillarReady then
			changed = true
		end
		self.pillarReady = pillarReady
	end

	if changed then
		self:UpdateProcIcon()
	end
end

function CompactRunes:UNIT_PET(event, unit)
	if unit and unit == "player" then
		if not _G.UnitExists("pet") then
			-- Ghoul is missing.
		end
	end
end

function CompactRunes:PLAYER_ENTERING_WORLD(event)
	for i = 1, 6 do
		self:UpdateRuneType(i)
		self:UpdateRuneCooldown(i)
	end
	self:UpdateRunicPower()
	self.playerGUID = _G.UnitGUID("player") or ""
end
